from __future__ import division
import numpy as np
from numpy import newaxis as na

from pyhsmm.basic.abstractions import GibbsSampling, MaxLikelihood, Distribution
from pyhsmm.basic.distributions import Regression, ARDRegression
from pyhsmm.util.stats import sample_mniw, sample_invwishart, sample_mn, \
        getdatasize

from util import AR_striding

class _ARMixin(object):
    @property
    def nlags(self):
        if not self.affine:
            return self.D_in // self.D_out
        else:
            return (self.D_in - 1) // self.D_out

    @property
    def D(self):
        return self.D_out

    def rvs(self,lagged_data):
        return super(_ARMixin,self).rvs(
                x=np.atleast_2d(lagged_data.ravel()),return_xy=False)

    def _get_statistics(self,data):
        return super(_ARMixin,self)._get_statistics(
                data=self._ensure_strided(data))

    def _get_weighted_statistics(self,data,weights):
        return super(_ARMixin,self)._get_weighted_statistics(
                data=self._ensure_strided(data),weights=weights)

    def log_likelihood(self,xy):
        return super(_ARMixin,self).log_likelihood(self._ensure_strided(xy))

    def _ensure_strided(self,data):
        if isinstance(data,np.ndarray):
            if data.shape[1] != self.D*(self.nlags+1):
                data = AR_striding(data,self.nlags)
            return data
        else:
            return [self._ensure_strided(d) for d in data]

class AutoRegression(_ARMixin,Regression):
    pass

class ARDAutoRegression(_ARMixin,ARDRegression):
    def __init__(self,M_0,**kwargs):
        blocksizes = [M_0.shape[0]]*(M_0.shape[1] // M_0.shape[0]) \
                + ([1] if M_0.shape[1] % M_0.shape[0] and M_0.shape[0] != 1 else [])
        super(ARDAutoRegression,self).__init__(
                M_0=M_0,blocksizes=blocksizes,**kwargs)




# distutils: language = c++
# distutils: extra_compile_args = -O2 -fopenmp -std=c++11 -DEIGEN_NO_MALLOC -DNDEBUG -w
# distutils: extra_link_args = -fopenmp
# cython: boundscheck = False

import numpy as np
cimport numpy as np

from libcpp.vector cimport vector
from libcpp cimport bool
from libc.stdint cimport int32_t, int64_t
from cython cimport floating, integral
from cython.parallel import prange

cdef extern from "messages.h":
    cdef cppclass dummy[Type]:
        dummy()
        Type resample_arhmm(
            int M, int T, int D, int nlags, bool affine,
            Type *pi_0, Type *A,
            Type *natparams, Type *normalizers,
            Type *data,
            Type *stats, int32_t *counts, int32_t *transcounts, int32_t *stateseq,
            Type *randseq, Type *alphan) nogil
        void initParallel()

def resample_arhmm(
        list pi_0s,
        list As,
        floating[:,:,::1] params,
        floating[::1] normalizers,
        list datas,
        list stateseqs,
        list randseqs,
        list alphans):
    cdef int i, j
    cdef dummy[floating] ref

    cdef int M = params.shape[0]   # number of states
    cdef int K = len(datas)        # number of sequences
    cdef int D = datas[0].shape[1] # dimension of data (unstrided)
    cdef bool affine = params.shape[2] % D
    cdef int nlags = (params.shape[2] - affine) / D - 1
    cdef int32_t[::1] Ts = np.array([d.shape[0] for d in datas]).astype('int32')

    cdef vector[int32_t*] stateseqs_v
    cdef int32_t[:] temp
    for i in range(K):
        temp = stateseqs[i]
        stateseqs_v.push_back(&temp[0])

    cdef vector[floating*] randseqs_v
    cdef floating[:] temp2
    for i in range(K):
        temp2 = randseqs[i]
        randseqs_v.push_back(&temp2[0])

    cdef vector[floating*] datas_v
    cdef floating[:,:] temp3
    for i in range(K):
        temp3 = datas[i]
        datas_v.push_back(&temp3[0,0])

    cdef vector[floating*] alphans_v
    cdef floating[:,:] temp4
    for i in range(K):
        temp4 = alphans[i]
        alphans_v.push_back(&temp4[0,0])

    cdef vector[floating*] As_v
    for i in range(K):
        temp4 = As[i]
        As_v.push_back(&temp4[0,0])

    cdef vector[floating*] pi_0s_v
    for i in range(K):
        temp2 = pi_0s[i]
        pi_0s_v.push_back(&temp2[0])

    # NOTE: 2*K for false sharing
    cdef int32_t[:,::1] ns = np.zeros((2*K,M),dtype='int32')
    cdef int32_t[:,:,::1] transcounts = np.zeros((2*K,M,M),dtype='int32')
    cdef floating[:,:,:,::1] stats
    cdef floating[::1] likes
    if floating is double:
        stats = np.zeros((2*K,M,params.shape[1],params.shape[2]),dtype='float64')
        likes = np.zeros(K,dtype='float64')
    else:
        stats = np.zeros((2*K,M,params.shape[1],params.shape[2]),dtype='float32')
        likes = np.zeros(K,dtype='float32')

    ref.initParallel()
    with nogil:
        for j in prange(K+1):
            if j != 0:
                i = j-1
                likes[i] = ref.resample_arhmm(
                        M,Ts[i],D,nlags,affine,
                        pi_0s_v[i],As_v[i],
                        &params[0,0,0],&normalizers[0],
                        datas_v[i],
                        &stats[2*i,0,0,0],&ns[2*i,0],&transcounts[2*i,0,0],
                        stateseqs_v[i], randseqs_v[i],alphans_v[i])

    allstats = []
    for statmat, n in zip(np.sum(stats,0),np.sum(ns,0)):
        xxT, yxT, yyT = statmat[:-D,:-D], statmat[-D:,:-D], statmat[-D:,-D:]
        allstats.append([yyT,yxT,xxT,n])

    return allstats, np.sum(transcounts,axis=0), np.asarray(likes)



from distutils.core import setup
from setuptools.command.build_ext import build_ext as _build_ext
from Cython.Build import cythonize
import numpy as np
from os.path import dirname, join, exists
from os import mkdir
from shutil import move
import tarfile
from urllib import urlretrieve
from glob import glob

# make dependency directory
if not exists('deps'):
    mkdir('deps')

# download Eigen if we don't have it in deps
eigenurl = 'http://bitbucket.org/eigen/eigen/get/3.2.6.tar.gz'
eigentarpath = join('deps', 'Eigen.tar.gz')
eigenpath = join('deps', 'Eigen')
if not exists(eigenpath):
    print('Downloading Eigen...')
    urlretrieve(eigenurl, eigentarpath)
    with tarfile.open(eigentarpath, 'r') as tar:
        tar.extractall('deps')
    thedir = glob(join('deps', 'eigen-eigen-*'))[0]
    move(join(thedir, 'Eigen'), eigenpath)
    print('...done!')

setup(
    name='autoregressive',
    version='0.1.0',
    description='Extension for switching vector autoregressive models with pyhsmm',
    author='Matthew James Johnson',
    author_email='mattjj@csail.mit.edu',
    url='https://github.com/mattjj/pyhsmm-autoregressive',
    license='GPL',
    packages=['autoregressive'],
    keywords=[
        'bayesian', 'inference', 'mcmc', 'time-series',
        'autoregressive', 'var', 'svar'],
    install_requires=[
        'Cython >= 0.20.1',
        'numpy', 'scipy', 'matplotlib', 'pybasicbayes', 'pyhsmm'],
    classifiers=[
        'Intended Audience :: Science/Research',
        'Programming Language :: Python',
        'Programming Language :: C++'],
    ext_modules=cythonize('**/*.pyx'),
    include_dirs=[np.get_include(), 'deps']
)



from __future__ import division
import numpy as np
from numpy.lib.stride_tricks import as_strided as ast

### striding data for efficient AR computations

def AR_striding(data,nlags):
    # I had some trouble with views and as_strided, so copy if not contiguous
    data = np.asarray(data)
    if not data.flags.c_contiguous:
        data = data.copy(order='C')
    if data.ndim == 1:
        data = np.reshape(data,(-1,1))
    sz = data.dtype.itemsize
    return ast(
            data,
            shape=(data.shape[0]-nlags,data.shape[1]*(nlags+1)),
            strides=(data.shape[1]*sz,sz))

def undo_AR_striding(strided_data,nlags):
    sz = strided_data.dtype.itemsize
    return ast(
            strided_data,
            shape=(strided_data.shape[0]+nlags,strided_data.shape[1]/(nlags+1)),
            strides=(strided_data.shape[1]/(nlags+1)*sz,sz))

### analyzing AR coefficient matrices

def canonical_matrix(A):
    # NOTE: throws away affine part
    D, nlags, _ = dimensions(A)
    mat = np.zeros((D*nlags,D*nlags))
    mat[:-D,D:] = np.eye(D*(nlags-1))
    mat[-D:,:] = A[:,:D*nlags]
    return mat

def eval_siso_transfer_function(A,from_idx,to_idx,freqs):
    D, _, _ = dimensions(A)
    assert 0 <= from_idx < D and 0 <= to_idx < D
    bigA = canonical_matrix(A)
    I = np.eye(bigA.shape[0])
    zs = np.exp(1j*np.array(freqs))
    return np.array(
            [np.linalg.inv(z*I-bigA)[-D:,-2*D:-D][to_idx,from_idx]
                for z in zs])

def is_affine(A):
    return bool(A.shape[1] % A.shape[0])

def is_stable(A):
    bigA = canonical_matrix(A)
    return np.all(np.abs(np.linalg.eigvals(bigA)) < 1.)

def dimensions(A):
    if is_affine(A):
        A = A[:,:-1]
    D, nlags = A.shape[0], A.shape[1] // A.shape[0]
    return D, nlags, is_affine(A)




from __future__ import division
import numpy as np
from matplotlib import pyplot as plt
from matplotlib.collections import LineCollection

import pyhsmm
from pyhsmm.models import _SeparateTransMixin
from pyhsmm.basic.distributions import Gaussian
from pyhsmm.util.general import cumsum
from pybasicbayes.util.general import blockarray

from util import AR_striding, undo_AR_striding


class _ARMixin(object):
    def __init__(self,init_emission_distn=None,**kwargs):
        super(_ARMixin,self).__init__(**kwargs)
        if init_emission_distn is None:
            init_emission_distn = \
                    Gaussian(nu_0=self.P+1,sigma_0=10*self.P*np.eye(self.P),
                        mu_0=np.zeros(self.P),kappa_0=1.)
        self.init_emission_distn = init_emission_distn

    def add_data(self,data,strided=False,**kwargs):
        strided_data = AR_striding(data,self.nlags) if not strided else data
        super(_ARMixin,self).add_data(data=strided_data,**kwargs)

    def _generate_obs(self,s):
        if s.data is None:
            # generating brand new data sequence
            data = np.zeros((s.T+self.nlags,self.D))

            if hasattr(self,'prefix'):
                data[:self.nlags] = self.prefix
            else:
                data[:self.nlags] = self.init_emission_distn\
                    .rvs().reshape(data[:self.nlags].shape)

            for idx, state in enumerate(s.stateseq):
                data[idx+self.nlags] = \
                    self.obs_distns[state].rvs(lagged_data=data[idx:idx+self.nlags])

            s.data = AR_striding(data,self.nlags)

        else:
            # filling in missing data
            data = undo_AR_striding(s.data,self.nlags)

            # TODO should sample from init_emission_distn if there are nans in
            # data[:self.nlags]
            assert not np.isnan(data[:self.nlags]).any(), "can't have missing data (nans) in prefix"

            nan_idx, = np.where(np.isnan(data[self.nlags:]).any(1))
            for idx, state in zip(nan_idx,s.stateseq[nan_idx]):
                data[idx+self.nlags] = \
                    self.obs_distns[state].rvs(lagged_data=data[idx:idx+self.nlags])

        return data

    ### Gibbs

    def resample_parameters(self):
        super(_ARMixin,self).resample_parameters()
        # self.resample_init_emission_distn()

    def resample_init_emission_distn(self):
        self.init_emission_distn.resample(
                [s.data[:self.nlags].ravel() for s in self.states_list])

    def _get_joblib_pair(self,s):
        return (undo_AR_striding(s.data,self.nlags),s._kwargs)

    ### convenient properties

    @property
    def nlags(self):
        return self.obs_distns[0].nlags

    @property
    def D(self):
        return self.obs_distns[0].D

    @property
    def P(self):
        return self.D*self.nlags

    @property
    def datas(self):
        return [undo_AR_striding(s.data,self.nlags) for s in self.states_list]

    ### plotting

    def _plot_2d_data_scatter(self,ax=None,state_colors=None,plot_slice=slice(None),update=False):
        # TODO doesn't play very nicely with plot_slice
        ax = ax if ax else plt.gca()

        artists = []
        for s, data in zip(self.states_list,self.datas):
            data = data[plot_slice]

            # scatter, like _HMMBase._plot_2d_data_scatter but pads the stateseq
            state_colors = state_colors if state_colors else self._get_colors()
            stateseq = np.concatenate((np.repeat(s.stateseq[0],self.nlags),s.stateseq))[plot_slice]
            colorseq = np.array([state_colors[state] for state in stateseq])
            if update and hasattr(s,'_data_scatter'):
                s._data_scatter.set_offsets(data[:,:2])
                s._data_scatter.set_color(colorseq)
            else:
                s._data_scatter = ax.scatter(data[:,0],data[:,1],c=colorseq,s=5)
            artists.append(s._data_scatter)

            # connect scatter points with line segments
            state_colors = self._get_colors(scalars=True)
            stateseq = np.concatenate((np.repeat(s.stateseq[0],self.nlags),s.stateseq[:-1]))[plot_slice]
            colorseq = np.array([state_colors[state] for state in stateseq])
            if update and hasattr(s,'_data_linesegments'):
                s._data_linesegments.set_array(colorseq)
            else:
                lc = s._data_linesegments = LineCollection(AR_striding(data,1).reshape(-1,2,2))
                lc.set_array(colorseq)
                lc.set_linewidth(0.5)
                ax.add_collection(lc)
            artists.append(s._data_linesegments)

        return artists

    def _plot_stateseq_data_values(self,s,ax,state_colors,plot_slice,update,data=None):
        data = undo_AR_striding(s.data,self.nlags)[plot_slice]
        stateseq = np.concatenate((np.repeat(s.stateseq[0],self.nlags),s.stateseq[:-1]))[plot_slice]
        colorseq = np.tile(np.array([state_colors[state] for state in stateseq]),data.shape[1])

        if update and hasattr(s,'_data_lc'):
            s._data_lc.set_array(colorseq)
        else:
            ts = np.arange(data.shape[0])
            segments = np.vstack(
                [AR_striding(np.hstack((ts[:,None], scalarseq[:,None])),1).reshape(-1,2,2)
                    for scalarseq in data.T])
            lc = s._data_lc = LineCollection(segments)
            lc.set_array(colorseq)
            lc.set_linewidth(0.5)
            ax.add_collection(lc)

        return s._data_lc


###################
#  model classes  #
###################

class ARHMM(_ARMixin,pyhsmm.models.HMM):
    pass


class ARWeakLimitHDPHMM(_ARMixin,pyhsmm.models.WeakLimitHDPHMM):
    pass


class ARHSMM(_ARMixin,pyhsmm.models.HSMM):
    pass


class ARWeakLimitHDPHSMM(_ARMixin,pyhsmm.models.WeakLimitHDPHSMM):
    pass


class ARWeakLimitStickyHDPHMM(_ARMixin,pyhsmm.models.WeakLimitStickyHDPHMM):
    pass


class ARWeakLimitHDPHSMMIntNegBin(_ARMixin,pyhsmm.models.WeakLimitHDPHSMMIntNegBin):
    pass


class ARWeakLimitGeoHDPHSMM(_ARMixin,pyhsmm.models.WeakLimitGeoHDPHSMM):
    pass


class ARHMMSeparateTrans(_ARMixin,pyhsmm.models.HMMSeparateTrans):
    pass


class ARWeakLimitHDPHMMSeparateTrans(_ARMixin,pyhsmm.models.WeakLimitHDPHMMSeparateTrans):
    pass


class ARWeakLimitHDPHSMMIntNegBinSeparateTrans(_ARMixin,pyhsmm.models.WeakLimitHDPHSMMIntNegBin):
    pass


class ARWeakLimitHDPHSMMDelayedIntNegBin(
        _ARMixin,
        pyhsmm.models.WeakLimitHDPHSMMDelayedIntNegBin):
    def resample_dur_distns(self):
        for state, distn in enumerate(self.dur_distns):
            distn.resample_with_censoring_and_truncation(
            data=
            [s.durations_censored[s.untrunc_slice][s.stateseq_norep[s.untrunc_slice] == state]
                - s.delays[state] for s in self.states_list],
            censored_data=
            [s.durations_censored[s.trunc_slice][s.stateseq_norep[s.trunc_slice] == state]
                - s.delays[state] for s in self.states_list])
        self._clear_caches()


class ARWeakLimitHDPHSMMDelayedIntNegBinSeparateTrans(
        _ARMixin,
        pyhsmm.models.WeakLimitHDPHSMMDelayedIntNegBinSeparateTrans):
    def resample_dur_distns(self):
        for state, distn in enumerate(self.dur_distns):
            distn.resample_with_censoring_and_truncation(
            data=
            [s.durations_censored[s.untrunc_slice][s.stateseq_norep[s.untrunc_slice] == state]
                - s.delays[state] for s in self.states_list],
            censored_data=
            [s.durations_censored[s.trunc_slice][s.stateseq_norep[s.trunc_slice] == state]
                - s.delays[state] for s in self.states_list])
        self._clear_caches()


### low-level code

class _HMMFastResamplingMixin(_ARMixin):
    _obs_stats = None
    _transcounts = []

    def __init__(self,dtype='float32',**kwargs):
        self.dtype = dtype
        super(_HMMFastResamplingMixin,self).__init__(**kwargs)

    def add_data(self,data,**kwargs):
        super(_HMMFastResamplingMixin,self).add_data(data.astype(self.dtype),**kwargs)

    def resample_states_slow(self,**kwargs):
        super(_HMMFastResamplingMixin,self).resample_states(**kwargs)

    def resample_states(self,**kwargs):
        from messages import resample_arhmm
        if len(self.states_list) > 0:
            stateseqs = [np.empty(s.T,dtype='int32') for s in self.states_list]
            params, normalizers = map(np.array,zip(*[self._param_matrix(o) for o in self.obs_distns]))
            stats, transcounts, loglikes = resample_arhmm(
                    [s.pi_0.astype(self.dtype) for s in self.states_list],
                    [s.trans_matrix.astype(self.dtype) for s in self.states_list],
                    params.astype(self.dtype), normalizers.astype(self.dtype),
                    [undo_AR_striding(s.data,self.nlags) for s in self.states_list],
                    stateseqs,
                    [np.random.uniform(size=s.T).astype(self.dtype) for s in self.states_list],
                    self.alphans)
            for s, stateseq, loglike in zip(self.states_list,stateseqs,loglikes):
                s.stateseq = stateseq
                s._normalizer = loglike

            self._obs_stats = stats
            self._transcounts = transcounts
        else:
            self._obs_stats = None
            self._transcounts = []

    def resample_obs_distns(self):
        if self._obs_stats is not None:
            for o, statmat in zip(self.obs_distns,self._obs_stats):
                o.resample(stats=statmat)
        else:
            super(_HMMFastResamplingMixin,self).resample_obs_distns()

    # TODO transcounts being lumped together in low-level code, needs separate
    # trans treatment. and negbin / hsmm treatment for that class.
    # def resample_trans_distn(self):
    #     self.trans_distn.resample(trans_counts=self._transcounts)

    @property
    def alphans(self):
        if not hasattr(self,'_alphans'):
            self._alphans = [np.empty((s.T,self.num_states),
                dtype=self.dtype) for s in self.states_list]
        return self._alphans

    @staticmethod
    def _param_matrix(o):
        D, A, sigma = o.D, o.A, o.sigma
        sigma_inv = np.linalg.inv(sigma)
        parammat =  -1./2 * blockarray([
            [A.T.dot(sigma_inv).dot(A), -A.T.dot(sigma_inv)],
            [-sigma_inv.dot(A), sigma_inv]
            ])
        normalizer = D/2*np.log(2*np.pi) + np.log(np.diag(np.linalg.cholesky(sigma))).sum()
        return parammat, normalizer


class FastARHMM(_HMMFastResamplingMixin,pyhsmm.models.HMM):
    pass

class FastARWeakLimitHDPHMM(_HMMFastResamplingMixin,pyhsmm.models.WeakLimitHDPHMM):
    pass

class FastARWeakLimitStickyHDPHMM(_HMMFastResamplingMixin,pyhsmm.models.WeakLimitStickyHDPHMM):
    pass

class FastARWeakLimitStickyHDPHMMSeparateTrans(_SeparateTransMixin,FastARWeakLimitStickyHDPHMM):
        _states_class = pyhsmm.internals.hmm_states.HMMStatesEigenSeparateTrans


#########################
#  changepoints models  #
#########################

class _ARChangepointsMixin(object):
    def add_data(self,data,changepoints,strided=False,**kwargs):
        nlags = self.nlags
        if sum(b-a for a,b in changepoints) != data.shape[0] - nlags:
            changepoints = [(max(a-nlags,0),b-nlags) for a,b in changepoints if b > nlags]
        super(_ARChangepointsMixin,self).add_data(
                data=data,changepoints=changepoints,**kwargs)


class ARWeakLimitHDPHSMMPossibleChangepoints(
        _ARChangepointsMixin,
        _ARMixin,
        pyhsmm.models.WeakLimitHDPHSMMPossibleChangepoints):
    pass


class ARWeakLimitHDPHSMMPossibleChangepointsSeparateTrans(
        _ARChangepointsMixin,
        _ARMixin,
        pyhsmm.models.WeakLimitHDPHSMMPossibleChangepointsSeparateTrans):
    pass
